Program DPendel;


(**************************************************************************)
(*                                                                        *)
(*  DPendel 2.0, a demo program for GNU Pascal and GRX which shows a      *)
(*  simulated film of a double pendulum.                                  *)
(*                                                                        *)
(*  Copyright (C) 1997  Peter Gerwinski, Essen, Germany                   *)
(*                      peter@gerwinski.de                                *)
(*                      http://www.gerwinski.de/~peter/                   *)
(*                                                                        *)
(*  This program is free software; you can redistribute it and/or modify  *)
(*  it under the terms of the GNU General Public License, version 2, as   *)
(*  published by the Free Software Foundation.  You can use parts of      *)
(*  this program in your own programs under the terms of the GNU Library  *)
(*  General Public License, version 2.                                    *)
(*                                                                        *)
(*  This program is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU General Public License for more details.                          *)
(*                                                                        *)
(*  You should have received a copy of the GNU General Public License     *)
(*  and the GNU Library General Public License along with this program;   *)
(*  if not, write to the Free Software Foundation, Inc., 675 Mass Ave,    *)
(*  Cambridge, MA 02139, USA.                                             *)
(*                                                                        *)
(**************************************************************************)


uses
  GRX;  (* Portable Graphics library *)


{$ignore-function-results}


Const
  m1 = 1.0;   (* Masses *)
  m2 = 1.0;
  l1 = 1.0;   (* Lengths *)
  l2 = 0.7;
  dt = 0.02;   (* Time step *)


(*** Vectors and Matrices ***)


Type
  VecR2 = record
    x, y: Real;
  end (* VecR2 *);

  MatR2 = record
    xx, xy, yx, yy: Real;
  end (* MatR2 *);

  TwoVectors = record
    phi, phidot: VecR2;
  end (* TwoVectors *);

  DglFuncPtr = ^Function ( Const State: TwoVectors ): TwoVectors;

  
(*** Variables for the actual calculation ***)


Var
  State, NewState: TwoVectors;


(*** Some global memory for the trace ***)


Const
  TracePoints = 1000;
  TraceColors = 13;


Var
  TracePoint: array [ 0..TracePoints - 1 ] of VecR2;
  Trace: Integer;
  TraceColor: array [ 0..TraceColors - 1 ] of Integer;


(*** VecR2 operations ***)


Operator + ( a, b: VecR2 ) = c: VecR2;

begin (* VecR2 + VecR2 *)
  c.x:= a.x + b.x;
  c.y:= a.y + b.y;
end (* VecR2 + VecR2 *);


Operator - ( a, b: VecR2 ) = c: VecR2;

begin (* VecR2 - VecR2 *)
  c.x:= a.x - b.x;
  c.y:= a.y - b.y;
end (* VecR2 - VecR2 *);


Operator * ( r: Real; a: VecR2 ) = c: VecR2;

begin (* Real * VecR2 *)
  c.x:= r * a.x;
  c.y:= r * a.y;
end (* Real * VecR2 *);


Operator * ( a: VecR2; r: Real ) = c: VecR2;

begin (* VecR2 * Real *)
  c.x:= a.x * r;
  c.y:= a.y * r;
end (* VecR2 * Real *);


Operator / ( a: VecR2; r: Real ) = c: VecR2;

begin (* VecR2 / Real *)
  c.x:= a.x / r;
  c.y:= a.y / r;
end (* VecR2 / Real *);


Operator * ( a, b: VecR2 ) = s: Real;

begin (* VecR2 * VecR2 *)
  s:= a.x * b.x + a.y * b.y;
end (* VecR2 * VecR2 *);


(*** MatR2 operations ***)


Operator + ( a, b: MatR2 ) = c: MatR2;

begin (* MatR2 + MatR2 *)
  c.xx:= a.xx + b.xx;
  c.xy:= a.xy + b.xy;
  c.yx:= a.yx + b.yx;
  c.yy:= a.yy + b.yy;
end (* MatR2 + MatR2 *);


Operator - ( a, b: MatR2 ) = c: MatR2;

begin (* MatR2 - MatR2 *)
  c.xx:= a.xx - b.xx;
  c.xy:= a.xy - b.xy;
  c.yx:= a.yx - b.yx;
  c.yy:= a.yy - b.yy;
end (* MatR2 - MatR2 *);


Operator * ( r: Real; a: MatR2 ) = c: MatR2;

begin (* Real * MatR2 *)
  c.xx:= r * a.xx;
  c.xy:= r * a.xy;
  c.yx:= r * a.yx;
  c.yy:= r * a.yy;
end (* Real * MatR2 *);


Operator * ( a: MatR2; r: Real ) = c: MatR2;

begin (* MatR2 * Real *)
  c.xx:= a.xx * r;
  c.xy:= a.xy * r;
  c.yx:= a.yx * r;
  c.yy:= a.yy * r;
end (* MatR2 * Real *);


Operator / ( a: MatR2; r: Real ) = c: MatR2;

begin (* MatR2 / Real *)
  c.xx:= a.xx / r;
  c.xy:= a.xy / r;
  c.yx:= a.yx / r;
  c.yy:= a.yy / r;
end (* MatR2 / Real *);


Operator * ( a: MatR2; b: VecR2 ) = c: VecR2;

begin (* MatR2 * MatR2 *)
  c.x:= a.xx * b.x + a.xy * b.y;
  c.y:= a.yx * b.x + a.yy * b.y;
end (* MatR2 * MatR2 *);


Operator * ( a, b: MatR2 ) = c: MatR2;

begin (* MatR2 * MatR2 *)
  c.xx:= a.xx * b.xx + a.xy * b.yx;
  c.xy:= a.xx * b.xy + a.xy * b.yy;
  c.yx:= a.yx * b.xx + a.yy * b.yx;
  c.yy:= a.yx * b.xy + a.yy * b.yy;
end (* MatR2 * MatR2 *);


(*** TwoVectors operations ***)


Operator + ( a, b: TwoVectors ) = c: TwoVectors;

begin (* TwoVectors + TwoVectors *)
  c.phi:= a.phi + b.phi;
  c.phidot:= a.phidot + b.phidot;
end (* TwoVectors + TwoVectors *);


Operator * ( r: Real; a: TwoVectors ) = c: TwoVectors;

begin (* Real * TwoVectors *)
  c.phi:= r * a.phi;
  c.phidot:= r * a.phidot;
end (* Real * TwoVectors *);


(*** Constructing vectors and matrices ***)


Function MakeVecR2 ( x, y: Real ) = c: VecR2;

begin (* MakeVecR2 *)
  c.x:= x;
  c.y:= y;
end (* MakeVecR2 *);


Function MakeMatR2 ( xx, xy, yx, yy: Real ) = c: MatR2;

begin (* MakeMatR2 *)
  c.xx:= xx;
  c.xy:= xy;
  c.yx:= yx;
  c.yy:= yy;
end (* MakeMatR2 *);


(*** Display everything on the screen ***)


Procedure DrawPendula ( phi1, phi2: Real; color: Integer );

Var
  x1, y1, x2, y2, x3, y3: Real;

begin (* DrawPendula *)
  x1:= GrScreenX div 2;
  y1:= GrScreenY div 2;
  x2:= x1 + l1 * GrScreenX / 5 * cos ( phi1 );
  y2:= y1 + l1 * GrScreenX / 5 * sin ( phi1 );
  GrLine ( Round ( x1 ), Round ( y1 ), Round ( x2 ), Round ( y2 ), color );
  x3:= x2 + l2 * GrScreenX / 5 * cos ( phi2 );
  y3:= y2 + l2 * GrScreenX / 5 * sin ( phi2 );
  GrLine ( Round ( x2 ), Round ( y2 ), Round ( x3 ), Round ( y3 ), color );
  with TracePoint [ Trace ] do
    if Color = GrWhite then
      begin
        GrPlot ( Round ( x1 ), Round ( y1 ), GrBlack );
        GrPlot ( Round ( x ), Round ( y ), GrBlack );
        x:= x3;
        y:= y3;
      end (* with *)
    else
      GrPlot ( Round ( x3 ), Round ( y3 ), TraceColor [ TraceColors - 1 ] );
end (* DrawPendula *);


Procedure DrawTrace;

Const
  ColorStep = TracePoints div TraceColors + 1;

Var
  i: Integer;

begin (* DrawTrace *)
  Trace:= ( Trace + 1 ) mod TracePoints;
  for i:= 0 to TracePoints - 1 do
    with TracePoint [ ( Trace + i ) mod TracePoints ] do
      GrPlot ( Round ( x ), Round ( y ), TraceColor [ i div ColorStep ] );
end (* DrawTrace *);


(*** Event handling ***)


(* Wait some MilliSeconds, but exit if an event happens. *)
(* Return whether a key was pressed.                     *)

Function WaitKeyPressed ( MilliSeconds: Integer ): Boolean;

Var
  Event: GrMouseEvent;

begin (* WaitKeyPressed *)
  GrMouseGetEventT ( Gr_m_KeyPress or Gr_m_NoPaint, @Event, MilliSeconds );
  WaitKeyPressed:= Event.flags and Gr_m_KeyPress <> 0;
end (* WaitKeyPressed *);


(*** Random numbers ***)


(* Derive a random number from the current second *)

Function TimeRandom: Real;

Var
  T: TimeStamp;

begin (* TimeRandom *)
  GetTimeStamp ( T );
  TimeRandom:= T.second / 30 - 1;
end (* TimeRandom *);


(*** Initialization and finalization ***)


Procedure Init;

Var
  i: Integer;

begin (* Init *)
  (* Initialize the trace *)
  for i:= 0 to TracePoints - 1 do
    TracePoint [ i ]:= MakeVecR2 ( -1, -1 );
  Trace:= 0;
  (* Set graphics mode *)
  GrSetMode ( gr_biggest_noninterlaced_graphics, 0, 0, 0, 0, 0 );
  (* Allocate custom colors *)
  TraceColor [ 0 ]:= GrAllocColor ( 0, 0, 31 );
  TraceColor [ 1 ]:= GrAllocColor ( 0, 0, 63 );
  TraceColor [ 2 ]:= GrAllocColor ( 0, 0, 127 );
  TraceColor [ 3 ]:= GrAllocColor ( 0, 0, 191 );
  TraceColor [ 4 ]:= GrAllocColor ( 0, 15, 255 );
  TraceColor [ 5 ]:= GrAllocColor ( 0, 31, 255 );
  TraceColor [ 6 ]:= GrAllocColor ( 0, 63, 255 );
  TraceColor [ 7 ]:= GrAllocColor ( 15, 127, 255 );
  TraceColor [ 8 ]:= GrAllocColor ( 31, 191, 255 );
  TraceColor [ 9 ]:= GrAllocColor ( 63, 255, 191 );
  TraceColor [ 10 ]:= GrAllocColor ( 127, 255, 127 );
  TraceColor [ 11 ]:= GrAllocColor ( 191, 255, 63 );
  TraceColor [ 12 ]:= GrAllocColor ( 255, 255, 31 );
  (* Initialize event handler *)
  GrMouseInit;
  (* Enable keyboard, disable mouse *)
  GrMouseEventEnable ( true, false );
end (* Init *);


Procedure Fini;

begin (* Fini *)
  (* Finalize event handler *)
  GrMouseUnInit;
  (* Go back to text mode *)
  GrSetMode ( gr_default_text, 0, 0, 0, 0, 0 );
end (* Fini *);


(*** Physics ***)


(* Calculate the derivative of the state with respect *)
(* to the time according to the Lagrange formalism.   *)
(*                                                    *)
(* These formulae have been calculated for this pro-  *)
(* gram by Dipl.-Phys. Anja Drewitz.  Thanks, Anja!  :*)

Function Lagrange ( Const State: TwoVectors ) = Derivative: TwoVectors;

Var
  prefactor, sindiff, cosdiff, nondiag: Real;
  phidot2: VecR2;
  Minv: MatR2;

begin (* Lagrange *)
  sindiff:= sin ( State.phi.x - State.phi.y );
  cosdiff:= cos ( State.phi.x - State.phi.y );
  phidot2.x:= sqr ( State.phidot.x );
  phidot2.y:= sqr ( State.phidot.y );
  nondiag:= - m2 * l1 * l2 * cosdiff;
  prefactor:= 1 / ( sqr ( l1 ) * sqr ( l2 ) * sqr ( m2 )
                    * ( 1 - sqr ( cosdiff ) + m1 / m2 ) );
  Minv:= prefactor * MakeMatR2 ( m2 * sqr ( l2 ), nondiag,
                                 nondiag, ( m1 + m2 ) * sqr ( l1 ) );
  with Derivative do
    begin
      phi:= State.phidot;
      phidot:= Minv * MakeVecR2 ( - m2 * l1 * l2 * phidot2.y * sindiff
                                  + ( m1 + m2 ) * l1 * cos ( State.phi.x ),
                                  m2 * l1 * l2 * phidot2.x * sindiff
                                  + m2 * l2 * cos ( State.phi.y ) );
    end (* with *);
end (* Lagrange *);


(*** Mathematics ***)


(* Approximate the state after the time step `dt' according *)
(* to the standard (explicit) Runge-Kutta algorithm.        *)

Function RungeKutta ( State: TwoVectors; dt: Real;
                      D: DglFuncPtr ) = NewState: TwoVectors;

Var
  k1, k2, k3, k4: TwoVectors;

begin (* RungeKutta *)
  k1:= D^ ( State );
  k2:= D^ ( State + 0.5 * ( dt * k1 ) );
  k3:= D^ ( State + 0.5 * ( dt * k2 ) );
  k4:= D^ ( State + dt * k3 );
  NewState:= State + dt * ( ( 1 / 6 ) * ( k1 + 2.0 * k2 + 2.0 * k3 + k4 ) );
end (* RungeKutta *);


(*** Main program ***)


begin
  Init;
  (* Initial position and speed *)
  with State do
    begin
      phi:= MakeVecR2 ( - pi / 2, 0.001 * TimeRandom - pi / 2 );
      phidot:= MakeVecR2 ( 0, 0 );
    end (* with *);
  repeat
    (* Calculation of new position and speed *)
    NewState:= RungeKutta ( State, dt, @Lagrange );
    DrawTrace;
    with State do
      DrawPendula ( phi.x, phi.y, GrBlack );
    with NewState do
      DrawPendula ( phi.x, phi.y, GrWhite );
    State:= NewState;
  until WaitKeyPressed ( 20 );
  Fini;
end.
