Program Fractals;


(**************************************************************************)
(*                                                                        *)
(*  Fractals 2.0, a demo program for GNU Pascal and its libraries which   *)
(*  graphically displays the following fractal sets:                      *)
(*    1) Julia set of the polynomial x**3 - (1+i),                        *)
(*    2) the Mandelbrot set.                                              *)
(*                                                                        *)
(*  Copyright (C) 1997  Peter Gerwinski, Essen, Germany                   *)
(*                      peter@gerwinski.de                                *)
(*                      http://www.gerwinski.de/~peter/                   *)
(*                                                                        *)
(*  This program is free software; you can redistribute it and/or modify  *)
(*  it under the terms of the GNU General Public License, version 2, as   *)
(*  published by the Free Software Foundation.  You can use parts of      *)
(*  this program in your own programs under the terms of the GNU Library  *)
(*  General Public License, version 2.                                    *)
(*                                                                        *)
(*  This program is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU General Public License for more details.                          *)
(*                                                                        *)
(*  You should have received a copy of the GNU General Public License     *)
(*  and the GNU Library General Public License along with this program;   *)
(*  if not, write to the Free Software Foundation, Inc., 675 Mass Ave,    *)
(*  Cambridge, MA 02139, USA.                                             *)
(*                                                                        *)
(**************************************************************************)


(* Fr Anja. *)


{$ignore-function-results}


uses
  GRX;


Type
  ColorFunc = ^Function ( x, y: Real ): Integer;


Const
  epsilon = 1E-15;

  LeftArrow  = $014B;
  RightArrow = $014D;
  UpArrow    = $0148;
  DownArrow  = $0150;


Var
  x0: Real value - 0.5;
  y0: Real value 0.0;
  size: Real value 2.0;
  Choice: Char;
  ScreenWidth, ScreenHeight, ScreenColors: Integer;
  Julia1, Julia2, Julia3: Integer;
  TheFractal: ColorFunc;
  PendingEvent: GrMouseEvent;
  finish: Boolean;


(*** GRX Keyboard support ***)


Function KeyPressed: Boolean;


begin (* KeyPressed *)
  if PendingEvent.flags = 0 then
    GrMouseGetEventT ( Gr_m_KeyPress or Gr_m_NoPaint, @PendingEvent, 0 );
  KeyPressed:= PendingEvent.flags and Gr_m_KeyPress <> 0;
end (* KeyPressed *);


Function ReadKey: Integer;


begin (* ReadKey *)
  repeat
  until KeyPressed;
  ReadKey:= PendingEvent.Key;
  if PendingEvent.Key = 27 then
    begin
      (* For Linux *)
      GrMouseGetEventT ( Gr_m_KeyPress or Gr_m_NoPaint, @PendingEvent, 1 );
      if ( PendingEvent.flags and Gr_m_KeyPress <> 0 )
         and ( PendingEvent.Key = ord ( '[' ) ) then
        begin
          GrMouseGetEventT ( Gr_m_KeyPress or Gr_m_NoPaint, @PendingEvent, 1 );
          case chr ( PendingEvent.Key ) of
            'A': ReadKey:= UpArrow;
            'B': ReadKey:= DownArrow;
            'C': ReadKey:= RightArrow;
            'D': ReadKey:= LeftArrow;
          end (* case *);
        end (* if *);
    end (* if *);
  PendingEvent.flags:= 0;
end (* ReadKey *);


(*** Mathematics ***)



(* The function for the Julia set *)

Function F ( x: Complex ): Complex;

begin (* F *)
  F:= x pow 3 - cmplx ( 1, 1 );
end (* F *);


(* Its derivative *)

Function dF ( x: Complex ): Complex;

begin (* dF *)
  dF:= cmplx ( 3, 0 ) * sqr ( x );
end (* dF *);


(* Newton's algorithm for finding zeros *)

Function Newton ( z: Complex ): Complex;

Var
  Fz, dFz: Complex;

begin (* Newton *)
  repeat
    Fz:= F ( z );
    dFz:= dF ( z );
    if abs ( dFz ) >= epsilon then
      z:= z - Fz / dFz;
  until ( abs ( Fz ) < epsilon )
     or ( abs ( dFz ) < epsilon )
     or ( abs ( z ) > 1E30 );
  Newton:= z;
end (* Newton *);


(*** Graphics ***)


(* Draw a solid square *)

Procedure Square ( x, y, a, color: Integer );

begin (* Square *)
  GrFilledBox ( x, y, x + a, y + a, color );
end (* Square *);


(* Draw the graph with increaing resolution *)

Procedure ColorGraph ( x0, x1, y0, y1: Real; Color: ColorFunc );

Var
  ix, iy, Step, c: Integer;
  x, y: Real;
  complete: Boolean;

begin (* ColorGraph *)
  Step:= 64;
  iy:= 0;
  while ( iy < ScreenHeight ) and not KeyPressed do
    begin
      y:= y1 + ( y0 - y1 ) * ( iy / ScreenHeight );
      ix:= 0;
      while ( ix < ScreenWidth ) and not KeyPressed do
        begin
          x:= x0 + ( x1 - x0 ) * ( ix / ScreenWidth );
          Square ( ix, iy, Step, GrWhite );
          c:= Color^ ( x, y );
          Square ( ix, iy, Step, c );
          inc ( ix, Step );
        end (* while *);
      inc ( iy, Step );
    end (* while *);
  Step:= Step div 2;
  while ( Step > 0 ) and not KeyPressed do
    begin
      iy:= 0;
      complete:= false;
      while ( iy < ScreenHeight ) and not KeyPressed do
        begin
          y:= y1 + ( y0 - y1 ) * ( iy / ScreenHeight );
          if complete then
            ix:= 0
          else
            ix:= Step;
          while ( ix < ScreenWidth ) and not KeyPressed do
            begin
              x:= x0 + ( x1 - x0 ) * ( ix / ScreenWidth );
              Square ( ix, iy, Step, GrWhite );
              c:= Color^ ( x, y );
              Square ( ix, iy, Step, c );
              if complete then
                inc ( ix, Step )
              else
                inc ( ix, 2 * Step );
            end (* while *);
          inc ( iy, Step );
          complete:= not complete;
        end (* while *);
      Step:= Step shr 1;
    end (* while *);
end (* ColorGraph *);


(*** The fractals ***)


(* Julia set *)

Function Julia ( x, y: Real ): Integer;

Var
  zero: Complex;

begin (* Julia *)
  zero:= Newton ( cmplx ( x, y ) );
  if Im ( zero ) > 0.5 then
    Julia:= Julia1
  else if Im ( zero ) > 0 then
    Julia:= Julia2
  else
    Julia:= Julia3;
end (* Julia *);


(* Mandelbrot set *)

Function Mandelbrot ( xx, yy: Real ): Integer;

Const
  tmax = 999;

Var
  t: Integer;
  x, y, z: Real;

begin (* Mandelbrot *)
  Mandelbrot:= GrBlack;
  t:= 0;
  if ( sqr ( xx + 0.24 ) + sqr ( yy ) > 0.24 )
   and ( sqr ( xx + 1 ) + sqr ( yy ) > 0.05 ) then
    begin
      x:= 0;
      y:= 0;
      repeat
        z:= sqr ( x ) - sqr ( y ) + xx;
        y:= 2 * x * y + yy;
        x:= z;
        inc ( t );
      until ( abs ( z ) > 100 ) or ( t > tmax );
      if t <= tmax then
        Mandelbrot:= t mod ( ScreenColors - 1 ) + 1;
    end (* if *);
end (* Mandelbrot *);


(*** Initialization and finalization ***)


Procedure PreInit;

begin (* PreInit *)
  GrSetMode ( gr_default_text, 0, 0, 0, 0, 0 );

  (* Initialize event handler *)

  GrMouseInit;

  (* Enable keyboard, disable mouse *)

  GrMouseEventEnable ( true, false );

  PendingEvent.flags:= 0;
end (* PreInit *);


Procedure Init ( Choice: Char );

begin (* Init *)
  GrSetMode ( gr_biggest_noninterlaced_graphics, 0, 0, 0, 0, 0 );
  ScreenWidth:= GrScreenX + 1;
  ScreenHeight:= GrScreenY + 1;
  ScreenColors:= GrNumColors;
  case Choice of
    '1': begin
           TheFractal:= @Julia;
           Julia1:= GrAllocColor ( 63, 255, 0 );
           Julia2:= GrAllocColor ( 255, 255, 0 );
           Julia3:= GrAllocColor ( 255, 63, 0 );
         end (* '1' *);
    '2': TheFractal:= @Mandelbrot;
  end (* case *);
end (* Init *);


Procedure Fini;

begin (* Fini *)
  (* Go back to text mode *)

  GrSetMode ( gr_default_text, 0, 0, 0, 0, 0 );

end (* Fini *);


Procedure PostFini;

begin (* PostFini *)
  (* Finalize event handler *)

  GrMouseUnInit;

end (* PostFini *);


(*** Main program ***)


begin
  PreInit;
  writeln;
  writeln ( 'arrows  -  walk' );
  writeln ( '+ -     -  zoom' );
  writeln ( 'ESC     -  quit' );
  writeln;
  writeln ( '1       -  Julia set of the polynomial x**3 - (1+i)' );
  writeln ( '2       -  Mandelbrot set' );
  writeln;
  repeat
    readln ( Choice );
  until Choice in [ '1', '2', #27 ];
  if Choice <> #27 then
    begin
      Init ( Choice );
      finish:= false;
      repeat
        ColorGraph ( x0 - 0.8 * size, x0 + 0.8 * size,
                     y0 - 0.6 * size, y0 + 0.6 * size, TheFractal );
        case ReadKey of
          ord ( '+' ): size:= sqrt ( 0.5 ) * size;
          ord ( '-' ): size:= sqrt ( 2 ) * size;
          LeftArrow:   x0:= x0 - 0.1 * size;
          RightArrow:  x0:= x0 + 0.1 * size;
          UpArrow:     y0:= y0 + 0.1 * size;
          DownArrow:   y0:= y0 - 0.1 * size;
          27:          finish:= true;
        end (* case *);
      until finish;
      Fini;
    end (* if *);
  PostFini;
end.
